#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <signal.h>

#include "pb-bonk.h"
#include "ui.h"

void doac (int sig);
void play (const char *filename);
void play_stop (void);

pid_t pb_bonk_pid;

void
help (void)
{
	printf ("playbonk version 0.2b\n\nUsage: playbonk (/some_dir/)\n\n");
	exit (-1);
}

UI *ui;

int
main (int argc, char **argv)
{
	signal (SIGCHLD, doac);

	if (argc == 1)
		ui = new UI ();
	else if (argc == 2)
		ui = new UI (argv[1]);
	else
		help ();
	
	while (1) switch (ui->input ()) {
	case UI_INPUT_PLAY:
		{
			char *tmp = ui->get_filename ();
			play (tmp);
			free (tmp);
		}
		break;
	case UI_INPUT_PLAY_STOP:
		play_stop ();
		break;
	case UI_INPUT_PLAY_NEXT:
		{
			char *tmp = ui->get_filename_next ();
			play (tmp);
			free (tmp);
		}
		break;
	case UI_INPUT_QUIT:
		delete ui;
		if (pb_bonk_pid != 0) {
			signal (SIGCHLD, SIG_IGN);
			kill (pb_bonk_pid, SIGKILL);
			wait (NULL);
			pb_bonk_pid = 0;
			signal (SIGCHLD, doac);
		}
		return 0;
	default: break;
	}
}

void
play (const char *filename)
{
	play_stop ();
	
	switch (pb_bonk_pid = fork ()) {
	case -1:
		abort ();
		break;
	case 0:
		play_file (strdup (filename));
		exit (0);
	default:
		ui->set_played_file ();
		ui->render_played_file ();
		break;
	}
}

void
play_stop (void)
{
	if (pb_bonk_pid != 0) {
		signal (SIGCHLD, SIG_IGN);
		kill (pb_bonk_pid, SIGKILL);
		wait (NULL);
		pb_bonk_pid = 0;
		signal (SIGCHLD, doac);
	}
	ui->set_played_file ("");
	ui->render_played_file ();
}

void 
doac (int sig)
{
	signal (SIGCHLD, doac);

	if (pb_bonk_pid != 0) {
		wait (NULL);
		pb_bonk_pid = 0;
	}

	ui->set_played_file ("");

	char *tmp = ui->played_file_next ();
	
	if (tmp == NULL) {
		ui->render_played_file ();
		return;
	}

	play (tmp);
	free (tmp);
}
