#include <stdlib.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "ui.h"

static void quit (int sig);

const char *file_pattern[] = { ".bnk", ".bonk", NULL };

UI::UI (const char *dir)
{
	char **dirs;
	char **files;
	
	/* Open a directory. */
	view_dir = new ViewDir (dir, file_pattern);

	dirs = (char **)view_dir->get_dirs ();
	files = (char **)view_dir->get_files ();
	dirs_size = view_dir->get_dirs_size ();
	files_size = view_dir->get_files_size ();

	signal (SIGINT, quit);
	initscr ();
	keypad (stdscr, TRUE);
	nonl ();
	cbreak ();
	noecho ();

	cols = COLS;
	rows = LINES;
	
	render_dir_name ();

	advance = true;
	render_advance ();

	is_stoped = false;

	played_file = NULL;
	played_file_list_nr = dirs_size + 1;
	render_played_file ();

	scroll_list = new UIScrollList (2, cols, 2, rows - 2);
	scroll_list->init_list_size (dirs_size + files_size + 1);
	scroll_list->init_item_add ("../");

	for (int i = 0; i < dirs_size; i++)
		scroll_list->init_item_add (dirs[i]);
	for (int i = 0; i < files_size; i++)
		scroll_list->init_item_add (files[i]);

	scroll_list->init_render_list ();
}

UI::~UI ()
{
	delete view_dir;
	delete scroll_list;
	endwin ();
}

UIInput
UI::input (void)
{
	const char *tmp;
	int ch;

	while (1) switch (ch = wgetch (stdscr)) {
	case KEY_DOWN:
		scroll_list->scroll_down ();
		break;
	case KEY_UP:
		scroll_list->scroll_up ();
		break;
	case KEY_PPAGE:
		scroll_list->scroll_page_up ();
		break;
	case KEY_NPAGE:
		scroll_list->scroll_page_down ();
		break;		
	case 13: /* ENTER */
		tmp = scroll_list->get_marked_item ();
		if (tmp[strlen (tmp) - 1] == '/') { 
			if (view_dir->change_working_dir (tmp) == true) {
				char **dirs;
				char **files;
				dirs = (char **)view_dir->get_dirs ();
				files = (char **)view_dir->get_files ();
				dirs_size = view_dir->get_dirs_size ();
				files_size = view_dir->get_files_size ();
				
				render_dir_name ();
				
				scroll_list->init_list_size (dirs_size +
							     files_size + 1);
				scroll_list->init_item_add ("../");	
				for (int i = 0; i < dirs_size; i++)
					scroll_list->init_item_add (dirs[i]);
				for (int i = 0; i < files_size; i++)
					scroll_list->init_item_add (files[i]);
				scroll_list->init_render_list ();
				
				is_stoped = true;
				played_file_list_nr = dirs_size + 1;
			}
		} else {
			played_file_list_nr = 
				scroll_list->get_marked_item_list_number ();
			is_stoped = false;
			return UI_INPUT_PLAY;
		}
		break;
	case KEY_LEFT:
		if (files_size != 0) {
			if (played_file_list_nr - 1 < dirs_size + 1)
				played_file_list_nr = dirs_size + files_size;
			else
				played_file_list_nr--;
			return UI_INPUT_PLAY_NEXT;
		}
		break;
	case KEY_RIGHT:
		if (files_size != 0) {
			if (played_file_list_nr + 1 > dirs_size + files_size)
				played_file_list_nr = dirs_size + 1;
			else
				played_file_list_nr++;
			return UI_INPUT_PLAY_NEXT;
		}
		break;
	case 330: /* del or backspace stopes. */
	case KEY_BACKSPACE:
	case KEY_DL:
		is_stoped = true;
		return UI_INPUT_PLAY_STOP;
		break;
	case 'q':
		return UI_INPUT_QUIT;
		break;
	case 'a':
		advance = !advance;
		render_advance ();
		break;
	case KEY_RESIZE: 
		{
			cols = COLS;
			rows = LINES;
			render_dir_name ();
			render_advance ();
			render_played_file ();
			
			/* Nothing is rendered here so we'll have to clean it our 
			 * self 
			 */
			for (int i = 0; i < cols; i++)
				mvaddch (rows - 2, i, ' ');
			for (int i = 1; i < rows - 1; i++)
				mvaddstr (i, 0, "  ");
			scroll_list->set_x_y (2, cols, 2, rows - 2);
		}
		break;
	default:
		break;
	};
}

char *
UI::get_filename (void) const
{
	char *filename;
	char *tmp;
	
	tmp = view_dir->get_working_dir ();

	filename = (char *)malloc (strlen (tmp) + 
				   strlen (scroll_list->get_marked_item ()) 
				   + 1);
	strcpy (filename, tmp);
	strcat (filename, scroll_list->get_marked_item ());
	free (tmp);
	return filename;
}

char *
UI::get_filename_next (void) const
{
	char *filename;
	char *tmp;
	
	tmp = view_dir->get_working_dir ();

	filename = (char *)malloc (strlen (tmp) + 
				   strlen (scroll_list->get_item 
					   (played_file_list_nr)) 
				   + 1);
	strcpy (filename, tmp);
	strcat (filename, scroll_list->get_item (played_file_list_nr));
	free (tmp);
	return filename;
}

void
UI::render_dir_name (void)
{
	/* Remove the old dir name. */
	for (int i = 0; i < cols - 4; i++)
		mvaddch (0, i, ' ');

	char *tmp = view_dir->get_working_dir_short ();
	mvaddstr(0, 0, tmp);
	free (tmp);
}

void
UI::render_advance (void)
{
	int x, y;
	getyx (stdscr, y, x);

	if (advance == true)
		mvaddstr (0, cols - 4, "a(*)");
	else
		mvaddstr (0, cols - 4, "a( )");       
	move (y, x);
}

void 
UI::set_played_file (const char *pf = NULL)
{
	if (pf == NULL) {
		if (played_file != NULL)
			free (played_file);
		if (played_file_list_nr > dirs_size + files_size)
			played_file = strdup ("");
		else 
			played_file = strdup (scroll_list->get_item 
					      (played_file_list_nr));
	} else {
		if (played_file != NULL)
			free (played_file);
		played_file = strdup (pf);
	}
}

char *
UI::played_file_next (void)
{
	if (played_file != NULL)
		free (played_file);
	played_file = NULL;

	if (files_size == 0 || advance == false || is_stoped != false) return NULL;
	
	if (played_file_list_nr + 1 > dirs_size + files_size)
		played_file_list_nr = dirs_size + 1;
	else
		played_file_list_nr++;

	char *filename;
	char *tmp;
	
	tmp = view_dir->get_working_dir ();
	
	filename = (char *)malloc (strlen (tmp) + 
				   strlen (scroll_list->get_item 
					   (played_file_list_nr)) + 1);
	strcpy (filename, tmp);
	strcat (filename, scroll_list->get_item (played_file_list_nr));
	free (tmp);
	played_file = strdup (scroll_list->get_item (played_file_list_nr));
	return filename;
}

void
UI::render_played_file (void)
{
	int x; 
	int y;

	getyx (stdscr, y, x);

	/* Remove the old dir name. */
	for (int i = 0; i < cols; i++)
		mvaddch (rows - 1, i, ' ');
	mvaddstr (rows - 1, 0, "playing (");
	if (played_file == NULL) {
		mvaddstr (rows - 1, 9, ")");
	} else {
		mvaddstr (rows - 1, 9, played_file);
		mvaddstr (rows - 1, 9 + strlen (played_file), ")");
	}
	move (y, x);
	refresh ();
}

static void
quit (int sig)
{
	endwin ();
	exit (0);
}
