#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>

#include "view-dir.h"

extern int errno;

ViewDir::ViewDir (const char *new_working_dir, const char **new_file_pattern)
{
	if (new_working_dir == NULL)
		new_working_dir = "~";

	dirs = NULL;
	files = NULL;

	file_pattern = new_file_pattern;

	if (strcmp (new_working_dir, "~") == 0)
		working_dir = strdup (getenv ("HOME"));
	else 
		working_dir = strdup (new_working_dir);	

	if (working_dir[strlen (working_dir) - 1] != '/') {
		char *foo = (char *)malloc (strlen (working_dir) + 2);
		strcpy (foo, working_dir);
		strcat (foo, "/");
		free (working_dir);
		working_dir = foo;
	}

	read_working_dir ();
}

ViewDir::~ViewDir ()
{
	if (working_dir != NULL)
		free (working_dir);
	if (dirs != NULL) {
		for (int i = 0; dirs[i] != NULL; i++)
			free (dirs[i]);
		free (dirs);
		dirs = NULL;
	}
	if (files != NULL) {
		for (int i = 0; files[i] != NULL; i++)
			free (files[i]);
		free (files);
		dirs = NULL;
	}
}

bool
ViewDir::change_working_dir (const char *new_working_dir)
{
	char *tmp = working_dir;
	if (strcmp (new_working_dir, "../") == 0) {
		if (strlen (working_dir) == 1)
			working_dir = strdup (working_dir);
		else {
			int i;
			i = strlen (working_dir);
			i--; /* skip '\0' */
			i--; /* skip '/' */
			while (i > 0 && working_dir[i] != '/') i--;			
			working_dir = (char *)malloc (i + 2);			
			for (int u = 0; u < i + 1; u++)
				working_dir[u] = tmp[u];
			working_dir[i + 1] = '\0';
		}
	} else {
		working_dir = (char *)malloc (strlen (working_dir) + 
					      strlen (new_working_dir) + 1);
		strcpy (working_dir, tmp);
		strcat (working_dir, new_working_dir);
	}
	
	if (read_working_dir () == false) {
		free (working_dir);
		working_dir = tmp;
		return false;
	}
	free (tmp);
	return true;
}

/* This one is useful because we need to know what to do if we can't open a directory
 * if it's the first time (e.g the user provided a directory at startup).
 * If it is we try to open the users home directory, and if that failes we die.
 * If it isn't the first time (e.g. the user tries to open a directory via playbonk)
 * we shouldn't do anything (if we can't open the dir in question that would be).
 */
static bool FIRST_TIME_READ_DIR = true;

static void bubble_sort (char **arr, int arr_size);

bool
ViewDir::read_working_dir (void)
{
	DIR *dir;
	struct dirent *dir_entry;
	struct stat file_type;

	dir = opendir (working_dir);

	if (errno == EMFILE ||   /* Too many file descriptors used by us. */
	    errno == ENFILE ||   /* Too many files open in the system. */
	    errno == ENOMEM)     /* Out of memory. */
		perror (strerror (errno));
	if (errno == EACCES ||   /* No permission */
	    errno == ENOENT ||   /* It doesn't exist */
	    errno == ENOTDIR) {  /* It's not a directory. */
		errno = 0;
		if (FIRST_TIME_READ_DIR == true) {
			printf ("\n** WARNING **: Couldn't open '%s'\n\n", working_dir);
			if (working_dir != NULL)
				free (working_dir);
			working_dir = strdup (getenv ("HOME"));
			dir = opendir (working_dir);
			if (errno == EACCES ||			    
			    errno == ENOENT ||
			    errno == ENOTDIR) {
				printf ("** ERROR **: Couldn't open '%s', your HOME directory\n\n", 
					working_dir);
				abort ();
			}
		} else
			return false;
	}

	FIRST_TIME_READ_DIR = false;

	if (dirs != NULL) {
		for (int i = 0; dirs[i] != NULL; i++)
			free (dirs[i]);
		free (dirs);
		dirs = NULL;
	}
	if (files != NULL) {
		for (int i = 0; files[i] != NULL; i++)
			free (files[i]);
		free (files);
		dirs = NULL;
	}

	readdir (dir); /* skip "." */
	readdir (dir); /* skip ".." */

	if (file_pattern == NULL) {
		int tmp_dirs_size = 0;
		int tmp_files_size = 0;
		
		while ((dir_entry = readdir (dir))) {
			if (dir_entry->d_name[0] == '.') continue;
			char *full_filename;
		        full_filename = (char *)malloc (1024);
			strcpy (full_filename, working_dir);
			strcat (full_filename, dir_entry->d_name);
			stat (full_filename, &file_type);
			if (S_ISDIR (file_type.st_mode) != 0)
				tmp_dirs_size++;
			else if (S_ISREG (file_type.st_mode) != 0)
				tmp_files_size++;				 
			free (full_filename);		
		}
		dirs = (char **)calloc (tmp_dirs_size + 1, sizeof (char *));
		if (tmp_files_size == 0) {
			files == NULL;
		} else
			files = (char **)calloc (tmp_files_size + 1, sizeof (char *));
		rewinddir (dir);
		readdir (dir); /* skip "." */
		readdir (dir); /* skip ".." */
		tmp_dirs_size = 0;
		tmp_files_size = 0;
		while ((dir_entry = readdir (dir))) {
			if (dir_entry->d_name[0] == '.') continue;
			char *full_filename;
		        full_filename = (char *)malloc (1024);
			strcpy (full_filename, working_dir);
			strcat (full_filename, dir_entry->d_name);
			stat (full_filename, &file_type);
			if (S_ISDIR (file_type.st_mode) != 0) {
				dirs[tmp_dirs_size] = (char *)malloc (strlen (dir_entry->d_name) + 2);
				strcpy (dirs[tmp_dirs_size], dir_entry->d_name);
				strcat (dirs[tmp_dirs_size], "/");
				tmp_dirs_size++;
			} else if (S_ISREG (file_type.st_mode) != 0) {
				files[tmp_files_size] = strdup (dir_entry->d_name);
				tmp_files_size++;				
			}
			free (full_filename);			
		}
		dirs[tmp_dirs_size] = NULL;
		if (tmp_files_size != 0)
			files[tmp_files_size] = NULL;
		dirs_size = tmp_dirs_size;
		files_size = tmp_files_size;
	} else {
		int tmp_dirs_size = 0;
		int tmp_files_size = 0;
		int file_match;
		
		while ((dir_entry = readdir (dir))) {
			if (dir_entry->d_name[0] == '.') continue;
			char *full_filename;
		        full_filename = (char *)malloc (1024);
			strcpy (full_filename, working_dir);
			strcat (full_filename, dir_entry->d_name);
			stat (full_filename, &file_type);
			if (S_ISDIR (file_type.st_mode) != 0)
				tmp_dirs_size++;
			else if (S_ISREG (file_type.st_mode) != 0) {
				file_match = false;
				for (int i = 0; 
				     file_match != true, file_pattern[i] != NULL; i++) {
					if (strlen (dir_entry->d_name) + 1 > 
					    strlen (file_pattern[i]) &&
					    strcasecmp (dir_entry->d_name +
							strlen (dir_entry->d_name) -
							strlen (file_pattern[i]), 
							file_pattern[i]) == 0)
						file_match = true;
				}
				if (file_match == true)
					tmp_files_size++;
			}
			free (full_filename);		
		}
		dirs = (char **)calloc (tmp_dirs_size + 1, sizeof (char *));
		if (tmp_files_size == 0) 
			files = NULL;
		else
			files = (char **)calloc (tmp_files_size + 1, sizeof (char *));
		rewinddir (dir);
		readdir (dir); /* skip "."  */
		readdir (dir); /* skip ".." */
		tmp_dirs_size = 0;
		tmp_files_size = 0;
		while ((dir_entry = readdir (dir))) {
			if (dir_entry->d_name[0] == '.') continue;
			char *full_filename;
		        full_filename = (char *)malloc (1024);
			strcpy (full_filename, working_dir);
			strcat (full_filename, dir_entry->d_name);
			stat (full_filename, &file_type);
			if (S_ISDIR (file_type.st_mode) != 0) {
				dirs[tmp_dirs_size] = (char *)malloc (strlen (dir_entry->d_name) + 2);
				strcpy (dirs[tmp_dirs_size], dir_entry->d_name);
				strcat (dirs[tmp_dirs_size], "/");
				tmp_dirs_size++;
			} else if (S_ISREG (file_type.st_mode) != 0) {
				file_match = false;
				for (int i = 0; 
				     file_match != true, file_pattern[i] != NULL; i++) {
					if (strlen (dir_entry->d_name) + 1 >
					    strlen (file_pattern[i]) &&
					    strcasecmp (dir_entry->d_name +
							strlen (dir_entry->d_name) -
							strlen (file_pattern[i]), 
							file_pattern[i]) == 0)
						file_match = true;
				}
				if (file_match == true) {
					files[tmp_files_size] = strdup (dir_entry->d_name);
					tmp_files_size++;
				}
			}
			free (full_filename);
		}
		dirs[tmp_dirs_size] = NULL;
		if (tmp_files_size != 0)
			files[tmp_files_size] = NULL;
		dirs_size = tmp_dirs_size;
		files_size = tmp_files_size;
	}

	closedir (dir);

	bubble_sort (dirs, dirs_size);
	bubble_sort (files, files_size);

	return true;
}

static void 
bubble_sort (char **arr, int arr_size)
{
	char *tmp;
	int have_sort;
	int have_top_sort;
	int have_top_sort_of;
	int arr_nr;
	int ch_nr;
	
	do {
		have_sort = 0;
		arr_nr = 0;
		while (arr_nr < arr_size - 1) {
			ch_nr = 0;
			
			do {
				have_top_sort = 0;
				have_top_sort_of = 0;
				
				if (arr[arr_nr][ch_nr] > arr[arr_nr+1][ch_nr]) {
					tmp = arr[arr_nr+1];
					arr[arr_nr+1] = arr[arr_nr];
					arr[arr_nr] = tmp;
					have_top_sort = 1;
				} else if (arr[arr_nr][ch_nr] < arr[arr_nr+1][ch_nr])
					have_top_sort_of = 1;
				ch_nr++;
			} while (arr[arr_nr][ch_nr] != '\0' &&
				 arr[arr_nr+1][ch_nr] != '\0' && 
				 have_top_sort_of != 1 && have_top_sort == 0);
			
			ch_nr = 0;

			if (have_top_sort == 1)
				have_sort = 1;

			arr_nr++;
		}
	} while (have_sort == 1);
}

char *
ViewDir::get_working_dir (void) const
{
	return strdup (working_dir);
}

char *
ViewDir::get_working_dir_short (void) const
{
	int i = strlen (working_dir);

	if (i == 1)
		return strdup (working_dir);
	i--; /* skip '\0' */
	i--; /* skip '/' */
	while (i > 0 && working_dir[i] != '/') i--;
	
	return strdup (working_dir + i + 1);
}
